/*
 * Copyright 2008 The Microlog project @sourceforge.net
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.microlog.midp.bluetooth;

import java.io.DataOutputStream;
import java.io.IOException;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;

import net.sf.microlog.core.Appender;
import net.sf.microlog.core.Level;
import net.sf.microlog.core.MicrologConstants;
import net.sf.microlog.core.appender.AbstractAppender;
import net.sf.microproperties.Properties;

/**
 * The <code>BluetoothSerialAppender</code> log using a Bluetooth serial
 * connection (btspp). It can be used in two different modes: try to locate the
 * Bluetooth logger server through regular Bluetooth lookup or by specifying the
 * exact url of the server. The latter version is useful for devices that fails
 * to lookup the server, for instance SonyEricsson P990i. If configured using a
 * property file the server url can be set using the property
 * <code>microlog.appender.BluetoothSerialAppender.serverUrl</code>
 * 
 * @author Johan Karlsson (johan.karlsson@jayway.se)
 * @author Jarle Hansen (hansjar@gmail.com)
 * @since 0.6
 */
public class BluetoothSerialAppender extends AbstractAppender {

	public static final String SERVER_URL_STRING = "microlog.appender.BluetoothSerialAppender.serverUrl";

	private String bluetoothFriendlyName = null;

	private String serverUrl;
	private StreamConnection connection;
	private DataOutputStream dataOutputStream;

	/**
	 * Default constructor, used when the device manages to locate the logging
	 * server on its own. The recommended way of using it.
	 */
	public BluetoothSerialAppender() {
	}

	/**
	 * Constructor that lets you specify the Bluetooth logger server instead of
	 * using the lookup service. This is useful if you have a device that fails
	 * to locate the Bluetooth logger server. Experienced with SonyEricsson
	 * P990i.
	 * 
	 * @param serverUrl
	 *            The server to connect to
	 */
	public BluetoothSerialAppender(String serverUrl) {
		this.serverUrl = serverUrl;
	}

	/**
	 * Constructor that lets you specify the Bluetooth Address of the server.
	 * This has several benefits. First it provides better performance than the
	 * default constructur, since a device search will normally take about 8-10
	 * seconds. Secondly it avoids hardcoding the server channel like the second
	 * constructor that receives a serverUrl String. The problem with hardcoding
	 * the channel is that you do not have any guarantee that it will be the
	 * same every time.
	 * 
	 * @param remoteDevice
	 *            The server to connect to, contains a Bluetooth Address
	 */
	public BluetoothSerialAppender(final BluetoothRemoteDevice remoteDevice) {
		final BluetoothServiceSearch bluetoothServiceSearch = BluetoothServiceSearch
				.newInstance();
		serverUrl = bluetoothServiceSearch.getLoggerServiceString(remoteDevice);
	}

	/**
	 * Clear the log. This has not affect for this appender.
	 */
	public void clear() {
		// Do nothing
	}

	public synchronized void shutdownLoggingService() throws IOException {
		if (dataOutputStream != null) {
			try {
				dataOutputStream.writeUTF("[STOP]");
			} catch (IOException io) {
				System.err
						.println("Failed to send [STOP] to the Bluetooth server, "
								+ io);
			} finally {
				close();
			}
		}
	}

	/**
	 * Close the log.
	 */
	public synchronized void close() throws IOException {
		if (dataOutputStream != null) {
			try {
				dataOutputStream.close();
			} catch (IOException e) {
				System.err
						.println("Failed to terminate the dataOutputStream in a controlled way."
								+ e);
			}
		}

		if (connection != null) {
			try {
				connection.close();
			} catch (IOException e) {
				System.err.println("Failed to close the log " + e);
			}
		}

		logOpen = false;
	}

	public synchronized void doLog(String clientID, String name, long time,
			Level level, Object message, Throwable t) {
		if (logOpen && formatter != null) {
			try {
				dataOutputStream.writeUTF(formatter.format(
						getBluetoothClientID(clientID), "", time, level,
						message, t));
				dataOutputStream.flush();
			} catch (IOException e) {
				System.err.println("Unable to log to the output stream. " + e);
			}
		}
	}

	/**
	 * Returns the clientID that will be used to identify the connected client.
	 * If the clientID is not set from the user it will try to use the Bluetooth
	 * friendly name instead.
	 * 
	 * @param clientID
	 *            the client id for the <code>BluetoothAppender</code>.
	 * 
	 * @return the id <code>String</code>
	 */
	private String getBluetoothClientID(final String clientID) {
		String bluetoothClientID = null;

		if (clientID == null || clientID.length() == 0) {
			bluetoothClientID = getBluetoothFriendlyName();
		}

		if (clientID != null || bluetoothClientID == null) {
			bluetoothClientID = clientID;
		}

		return bluetoothClientID;
	}

	/**
	 * Tries to retrieve the Bluetooth friendly name. If this is successful the
	 * first time the same name will be returned each time.
	 * 
	 * @return the Bluetooth friendly name, that is the name specified by the
	 *         user.
	 */
	private String getBluetoothFriendlyName() {
		if (bluetoothFriendlyName == null) {
			try {
				bluetoothFriendlyName = LocalDevice.getLocalDevice()
						.getFriendlyName();
			} catch (BluetoothStateException bse) {
				System.err
						.println("Unable to get the Bluetooth friendly name, "
								+ bse);
			}
		}

		return bluetoothFriendlyName;
	}

	/**
	 * Open the log, i.e. open the Bluetooth connection to the log server.
	 */
	public synchronized void open() throws IOException {
		try {
			String connectionString = null;
			if (serverUrl == null) {
				// Try to find a logger server using Bluetooth discovery
				// Retrieve the connection string to connect to
				// the server
				LocalDevice local = LocalDevice.getLocalDevice();
				DiscoveryAgent agent = local.getDiscoveryAgent();

				connectionString = agent.selectService(new UUID(
						MicrologConstants.DEFAULT_BT_UUID_STRING, false),
						ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);
			} else {
				connectionString = serverUrl;
			}

			if (connectionString != null) {
				try {
					connection = (StreamConnection) Connector
							.open(connectionString);
					dataOutputStream = connection.openDataOutputStream();

					logOpen = true;
				} catch (IOException e) {
					System.err
							.println("Failed to connect to the Bluetooth log server with connection string "
									+ connectionString + ' ' + e);
				}
			} else {
				System.err.println("Did not find any Microlog service.");
			}
		} catch (BluetoothStateException e) {
			System.err
					.println("Failed to connect to the Bluetooth log server. "
							+ e);
		}
	}

	/**
	 * Get the size of the. Always returns <code>Appender.SIZE_UNDEFINED</code>.
	 */
	public long getLogSize() {
		return Appender.SIZE_UNDEFINED;
	}

	/**
	 * Configure the BluetoothSerialAppender.
	 * <p>
	 * If the device for some reason fails to locate a bluetooth logger server
	 * the connection url can set hard using the property
	 * <code>microlog.appender.BluetoothSerialAppender.serverUrl</code>.
	 * 
	 * @param properties
	 *            Properties to configure with
	 */
	public synchronized void configure(Properties properties) {

		// Set the record store name from Properties
		String serverUrlString = properties.getProperty(SERVER_URL_STRING);
		if (serverUrlString != null && serverUrlString.length() > 0) {
			serverUrl = serverUrlString;
		}
	}

	/**
	 * @see Appender#getPropertyNames()
	 */
	public String[] getPropertyNames() {
		return null;
	}

	/**
	 * @see Appender#setProperty(String, String)
	 */
	public void setProperty(String name, String value)
			throws IllegalArgumentException {
		// TODO Auto-generated method stub
	}

}
